%% StoIHT.m
function [w myapprox CPUTime myidx] = f_StoIHT(A, u, s, blocksize, gamma, p, tol, maxIter)
% Description: Stochastic Iterative Hard Thresholding (StoIHT) algorithm 
% to approximate the vector x from measurements u = A*x.
%
% Usage: [w myapprox CPUTime myidx] = f_StoIHT(A, u, s, blocksize, gamma, p, tol, maxIter)
%  Inputs:
%         A = measurement matrix
%         u = (noisy) measurements
%         s = sparsity level
% blocksize = block size
%     gamma = step size
%         p = vector of probabilities of length size(A,1)
%       tol = tolerance used to exit loop
%   maxIter = maximum number of iterations
%  Outputs:
%         w = s-sparse approximation of target signal
%  myapprox = matrix holding column vectors of signal 
%            approximation at each iteration
%   CPUTime = elasped CPU time after each iteration
%     myidx = matrix holding column vectors of identified
%             indices at each iteration
%         

% Initialize variables
cpuinit = cputime;
d = size(A,2);           % signal dimension
w = zeros(d,1);          % initial signal approximation
M = size(A,1);           % number of measurements
k = 0;                   % loop index
done = 0;                % while loop flag
nb = floor(M/blocksize); % number of blocks


while (~done)
   k = k+1;

   % Randomize
   ii = ceil(rand*nb); 
   block = (blocksize*(ii-1)+1) : (blocksize*ii);
   
   % Proxy
   gradfi = -2*A(block,:)'*(u(block)-A(block,:)*w);
   b = w-gamma/(nb*p(ii))*gradfi;
   
   % Estimate
   [b_sort ix_b] = sort(abs(b),'descend');
   w = zeros(d,1);
   w(ix_b(1:s)) = b(ix_b(1:s)); 
   
   % Keep track of approximations at each iteration
   myapprox(1:d,k) = w;
   % Keep track of identified indices at each iteration
   myidx(1:s,k)    = ix_b(1:s);
   
   % Check exit criteria
   if (norm(u-A*w,2) <= tol || k >= maxIter)
       done = 1;
   end
   
   CPUTime(k) = cputime-cpuinit;
   
end

end

