%% StoIHT.m
function [w_out myapprox CPUTime myidx done] = f_StoIHT_iteration_tally(A, u, s, blocksize, gamma, p, tol, maxIter, w, tally, k)
% Description: Stochastic Iterative Hard Thresholding (StoIHT) algorithm 
% to approximate the vector x from measurements u = A*x.
%
% Usage: [w myapprox CPUTime myidx done] = f_StoIHT_iteration_tally(A, u, s, blocksize, gamma, p, tol, maxIter, w, k)
%  Inputs:
%         A = measurement matrix
%         u = (noisy) measurements
%         s = sparsity level
% blocksize = block size
%     gamma = step size
%         p = vector of probabilities of length size(A,1)
%       tol = tolerance used to exit loop
%   maxIter = maximum number of iterations
%         w = current signal estimate
%     tally = input tally locations
%         k = current iteration
%  Outputs:
%     w_out = approximation to be used in next iteration
%  myapprox = s-sparse approximation of signal
%   CPUTime = elasped CPU time after each iteration
%     myidx = identified indices at this iteration
%      done = flag if exit criteria has be achieved
%         

% Initialize variables
cpuinit = cputime;
d = size(A,2);           % signal dimension
M = size(A,1);           % number of measurements
done = 0;                % while loop flag
nb = floor(M/blocksize); % number of blocks

% Randomize
ii = ceil(rand*nb); 
block = (blocksize*(ii-1)+1) : (blocksize*ii);

% Proxy
gradfi = -2*A(block,:)'*(u(block)-A(block,:)*w);
b = w-gamma/(nb*p(ii))*gradfi;

% Estimate
[b_sort ix_b] = sort(abs(b),'descend');
w_out = zeros(d,1);
w_out(ix_b(1:s)) = b(ix_b(1:s)); % project onto largest s elements
w_out(tally) = b(tally);         % project onto tally

% Keep track of approximations at each iteration
myapprox = w_out;
% Keep track of identified indices at each iteration
myidx    = ix_b(1:s);

% Check exit criteria
if (norm(u-A*w_out,2) <= tol || k >= maxIter)
   done = 1;
end

CPUTime = cputime-cpuinit;

end

