%% Initialize
clear; 
close all;
clc;
rng('default') % seed the random number generator to produce a predictable sequence of numbers

%% Declare Algorithm Parameters
n         = 500;                % number of trials
d         = 1000;               % signal dimension
s         = 20;                 % sparsity level of signal
m         = 300;                % number of measurements
tol_alg   = 1e-7;               % tolerance to quit in algorithm
maxEpochs = 100;                % maximum epochs in algorithms
gamma     = 1;                  % step size for iterative hard thresholding algorithms (IHT and StoIHT)
ncoreslist    = [2 4 6 10 20 30 40 50];             % number of processors/cores to simulate
nslowcores =    [0 0 0 0  0  0  0  0];              % number of slow cores (should have same dimension as ncoreslist)
slowrate   =    [0 0 0 0  0  0  0  0];              % slow cores get updated every given number of iterations (i.e., 2 means the slow cores get updated every other iteration) (should have same dimension as ncoreslist)

%% Run Experiments
for is = 1:length(s) % for each sparsity level
    for im = 1:length(m) % for each measurement value
        blocksize  = min(m(im),s(is));% block size
        nb = floor(m(im)/blocksize); % number of blocks
        p  = ones(1,nb)*1/nb; % set probabilities
        maxIter = nb*maxEpochs;  % maximum number of iterations
        for in = 1:n % for each trial
            
            % Generate s(is)-sparse signal
            x                  = zeros(d,1);
            rand_perm          = randperm(d);
            support_indices    = rand_perm(1:s(is));
            x(support_indices) = sign(randn(length(support_indices),1)); % Bernoulli signal
            trueSupport{is,im}(1:s(is),in) = support_indices;
            
            % Generate measurement matrix
            A = randn(m(im),d)/sqrt(m(im)); % Gaussian measurement matrix
            
            % Compute measurements
            y = A*x; % measurements
            
            %%% Run Standard Algorithms %%%
            
            %%% Run Standard StoIHT %%%
            [w myapprox CPUTime myidx] = f_StoIHT(A, y, s(is), blocksize, gamma, p, tol_alg, maxIter);

            stoiht_Seq(is,im,in).lastk = size(myapprox,2);
            stoiht_Seq(is,im,in).errPerIter = sqrt(sum((repmat(x,1,size(myapprox,2))-myapprox).^2,1));
            stoiht_Seq(is,im,in).lasterr = norm(x-w);
            stoiht_Seq(is,im,in).cputime = CPUTime;    

            stoiht_Seq(is,im,in).tally_idx = myidx;
            
            %%% Simulate Algorithms Asynchronously on 'ncores' cores using a shared tally %%%
            
            for ic = 1:length(ncoreslist)
                
                ncores = ncoreslist(ic);
                    
                % Initialize core data
                Core = [];
                for ii = 1:ncores
                   Core(ii).w       = zeros(d,1);
                   Core(ii).done    = 0;
                   Core(ii).k       = 0;
                end
                k_shared = 0; % shared GROUP iteration counter amongst cores
                % support tally counter for asynchronous method
                % Contains each core's most recent guess only!
                stoiht_Async(is,im,in,ic).tally = zeros(d,1); 

                while ( sum([Core(:).done]) == 0 )
                    k_shared = k_shared + 1;

                    % identify s largest (and nonzero) indices from the tally
                    [tally_sort ix_tally] = sort(stoiht_Async(is,im,in,ic).tally,'descend');
                    T = ix_tally( 1:min(nnz(tally_sort),s(is)) );

                    stoiht_Async(is,im,in,ic).tally_acc(k_shared) = length(intersect(T,support_indices))/s(is);

                    stoiht_Async(is,im,in,ic).tally = zeros(d,1); % reset tally

                    for ind = 1:ncores

                        % check for slow cores (cores 1:nslowcores are slow)
                        if (ind <= nslowcores(ic) && (k_shared-1)/slowrate(ic) ~= round((k_shared-1)/slowrate(ic)))
                            % if this is a slow core, it doesn't get updated this iteration
                            stoiht_Async(is,im,in,ic).errPerIter(ind,k_shared) = stoiht_Async(is,im,in,ic).errPerIter(ind,k_shared-1);
                            stoiht_Async(is,im,in,ic).myidx{ind}(1:s(is),k_shared) = stoiht_Async(is,im,in,ic).myidx{ind}(1:s(is),k_shared-1);
                            %stoiht_Async(is,im,in,ic).tally(stoiht_Async(is,im,in,ic).myidx{ind}(1:s(is),k_shared)) = stoiht_Async(is,im,in,ic).tally(stoiht_Async(is,im,in,ic).myidx{ind}(1:s(is),k_shared)) + 1; % always add 1 to the tally
                            stoiht_Async(is,im,in,ic).tally(stoiht_Async(is,im,in,ic).myidx{ind}(1:s(is),k_shared)) = stoiht_Async(is,im,in,ic).tally(stoiht_Async(is,im,in,ic).myidx{ind}(1:s(is),k_shared)) + Core(ind).k;

                            continue;
                        end

                        % Run StoIHT iteration
                        Core(ind).k = Core(ind).k + 1;
                        [w_out myapprox CPUTime myidx done] = f_StoIHT_iteration_tally(A, y, s(is), blocksize, gamma, p, tol_alg, maxIter, Core(ind).w, T, Core(ind).k);

                        Core(ind).done = done;
                        Core(ind).w    = w_out;

                        if (Core(ind).done)
                            stoiht_Async(is,im,in,ic).lastk = k_shared;
                            stoiht_Async(is,im,in,ic).lasterr = norm(x-myapprox);
                        end
                        stoiht_Async(is,im,in,ic).errPerIter(ind,k_shared) = norm(x-myapprox);
                        stoiht_Async(is,im,in,ic).myidx{ind}(1:s(is),k_shared) = myidx;

                        % update the tally
                        %stoiht_Async(is,im,in,ic).tally(myidx) = stoiht_Async(is,im,in,ic).tally(myidx) + 1; % always add 1 to the tally
                        stoiht_Async(is,im,in,ic).tally(myidx) = stoiht_Async(is,im,in,ic).tally(myidx) + Core(ind).k;
                    end

                end
            
            end
            
            disp(['Trial ' num2str(in)]);
        end
        
        % Command window output to display progress
        disp(['Completed sparsity ' num2str(s(is)) ' with ' num2str(m(im)) ' measurements']);
        
    end
end

%% Plot Results 

% Calculate average number of iterations to exit
is = 1;
im = 1;
mean_stoiht_Seq = mean([stoiht_Seq(is,im,:).lastk]);
std_stoiht_Seq = std([stoiht_Seq(is,im,:).lastk]);
for ic = 1:length(ncoreslist)
    mean_stoiht_Async(ic) = mean([stoiht_Async(is,im,:,ic).lastk]);
    std_stoiht_Async(ic) = std([stoiht_Async(is,im,:,ic).lastk]);
end

figure; hold on;
hf = fill([ncoreslist fliplr(ncoreslist)], [mean_stoiht_Seq*ones(1,length(ncoreslist))-std_stoiht_Seq fliplr(mean_stoiht_Seq*ones(1,length(ncoreslist))+std_stoiht_Seq)],'r','edgecolor','none');
alpha(hf,0.2)
hf = fill([ncoreslist fliplr(ncoreslist)], [mean_stoiht_Async-std_stoiht_Async fliplr(mean_stoiht_Async+std_stoiht_Async)],'b','edgecolor','none');
alpha(hf,0.2)
plot(ncoreslist,mean_stoiht_Seq*ones(1,length(ncoreslist)),'--r','linewidth',2);
plot(ncoreslist,mean_stoiht_Async,'-o','linewidth',2);
xlabel('Number of Cores','fontsize',16);
ylabel('Number of Time Steps','fontsize',16);
set(gca,'fontsize',14);
title({'StoIHT: Number of Iterations to Exit';['s = ' num2str(s(is)) ', m = ' num2str(m(im)) ', Trials = ' num2str(n)]},'fontsize',16);
%grid on;
hl = legend('Standard, +/- 1 Standard Deviation','Async, +/- 1 Standard Deviation','Standard Mean','Async Mean');
PatchInLegend = findobj(hl,'type','patch');
set(PatchInLegend,'facea',0.2);
